;; Generated package description from minimail.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "minimail" "0.3.0.20260404.123801" "Simple, non-blocking IMAP email client" '((emacs "30.1")) :commit "223cab1aa82045cdddd3d7ec284ee699597390e2" :authors '(("Augusto Stoffel" . "arstoffel@gmail.com")) :maintainer '("Augusto Stoffel" . "arstoffel@gmail.com") :keywords '("mail") :url "https://codeberg.org/astoff/minimail")
